""" Common Errors """

REST_REQUEST_DATA		= 100
REST_CONTENT_TYPE		= 101
REST_AUTHORIZATION		= 102
REST_LIST_TO_LONG		= 103

SERVICE_ACTION			= 200
SERVICE_STATUS			= 201
SERVICE_CONTENT_TYPE	= 202
SERVICE_UNREACHABLE		= 203
SERVICE_NOT_REGISTERED	= 204
SERVICE_NO_SUCH_NOUN	= 205
SERVICE_INTERNAL_KEY	= 206
SERVICE_CRASHED			= 207
SERVICE_NO_DATA			= 208
SERVICE_NO_SESSION		= 209